# sb::Plugin::sbTextExtension
# == written by T.Otani <takuya.otani@gmail.com> ===
# == Copyright (C) 2008 SimpleBoxes/SerendipityNZ ==

use strict;

if ( !defined $sb::VERSION
  or $sb::VERSION =~ /^2.(\d\d)[BRD]$/i )
{
	die("This plugin works under Serene Bach 3.* only.\n");
}

# ======================================
package sb::Plugin::sbTextExtension;
use sb::Plugin ();
# --------------------------------------
sb::Plugin->register_plugin(
	'lang' => {
		'ja' => 'utf8',
		'en' => 'ascii',
	},
	'text' => {
		'type'    => 'sbtext extenstion',
		'name'    => 'sbtext extenstion',
		'text'    => 'sbtext extenstion',
		'author'  => 'takkyun',
		'detail'  => 'http://serenebach.net/',
		'version' => version(),
	},
	'file' => undef,
	'data' => undef,
);
sub version
{
	return '0.00';
	# 0.00 [2008/12/04] generated
}
# ======================================
package sbtext::Plugin::html;
sub version
{
	return '0.00';
	# 0.00 [2008/12/04] generated
}
sub range
{
	my $class = shift;
	my $param = &_detitize(shift);
	my $obj = shift;
	my $comreg = '<!(?:--[^-]*-(?:[^-]+-)*?-(?:[^>-]*(?:-[^>-]+)*?)??)*(?:>|$(?!\n)|--.*$)';
	$param =~ s/(<[\w\/]+?[^>]*?>)/'<em class="keyword">' . &_entitize($1) . '<\/em>'/eg;
	$param =~ s/($comreg)/'<span class="comment">' . &_entitize($1) . '<\/span>'/eg;
	return '<pre><code class="html">' . $param . '</code></pre>';
}
sub _entitize
{
	return sb::Plugin::sbTextParser::_entitize(@_);
}
sub _detitize
{
	return sb::Plugin::sbTextParser::_detitize(@_);
}
sub syntax
{
	my $class = shift;
	my $type = shift;
	return '>|html|\\nstring\\n||<' if ($type eq 'range');
	return;
}
# ======================================
package sbtext::Plugin::css;
sub version
{
	return '0.00';
	# 0.00 [2008/12/04] generated
}
sub range
{
	my $class = shift;
	my $param = shift;
	my $obj = shift;
	my $keyword = $class->_keyword;
	$param =~ s/(\".*?\")/<span class="quote">$1<\/span>/g;
	$param =~ s/(\'.*?\')/<span class="quote">$1<\/span>/g;
	$param =~ s/$keyword:/<em class="keyword">$1<\/em>:/g;
	$param =~ s/(\/\*.*?\*\/)/<span class="comment">$1<\/span>/g;
	return '<pre><code class="css">' . $param . '</code></pre>';
}
sub _keyword
{
	return '(azimuth|background|background\-attachment|background\-color|background\-image|background\-position|background\-repeat|border|border\-collapse|border\-color|border\-spacing|border\-style|border\-top|border\-right|border\-bottom|border\-left|border\-top\-color|border\-right\-color|border\-bottom\-color|border\-left\-color|border\-top\-style|border\-right\-style|border\-bottom\-style|border\-left\-style|border\-top\-width|border\-right\-width|border\-bottom\-width|border\-left\-width|border\-width|bottom|caption\-side|clear|clip|color|content|counter\-increment|counter\-reset|cue|cue\-after|cue\-before|cursor|direction|display|elevation|empty\-cells|float|font|font\-family|font\-size|font\-size\-adjust|font\-stretch|font\-style|font\-variant|font\-weight|height|left|letter\-spacing|line\-height|list\-style|list\-style\-image|list\-style\-position|list\-style\-type|margin|margin\-top|margin\-right|margin\-bottom|margin\-left|marker\-offset|marks|max\-height|max\-width|min\-height|min\-width|orphans|outline|outline\-color|outline\-style|outline\-width|overflow|padding|padding\-top|padding\-right|padding\-bottom|padding\-left|page|page\-break\-after|page\-break\-before|page\-break\-inside|pause|pause\-after|pause\-before|pitch|pitch\-range|play\-during|position|quotes|richness|right|size|speak|speak\-header|speak\-numeral|speak\-punctuation|speech\-rate|stress|table\-layout|text\-align|text\-decoration|text\-indent|text\-shadow|text\-transform|top|unicode\-bidi|vertical\-align|visibility|voice\-family|volume|white\-space|widows|width|word\-spacing|z\-index)';
}
sub syntax
{
	my $class = shift;
	my $type = shift;
	return '>|css|\\nstring\\n||<' if ($type eq 'range');
	return;
}
# ======================================
package sbtext::Plugin::js;
sub version
{
	return '0.00';
	# 0.00 [2008/12/04] generated
}
sub range
{
	my $class = shift;
	my $param = shift;
	my $obj = shift;
	my $keyword = $class->_keyword;
	$param =~ s/(\&qout;.*?\&qout;)/&sbz;$1<\/span>/g;
	$param =~ s/(\'.*?\')/&sbz;$1<\/span>/g;
	$param =~ s/$keyword/<em class="keyword">$1<\/em>/g;
	$param =~ s/&sbz;/<span class="qoute">/g;
	$param =~ s/(\/\/.+)\n/<span class="comment">$1<\/span>\n/g;
	$param =~ s/(\/\*.*?\*\/)/<span class="comment">$1<\/span>/g;
	return '<pre><code class="js">' . $param . '</code></pre>';
}
sub _keyword
{
	return '(window|document|location|navigator|history|abstract|boolean|break|byte|case|catch|char|class|const|continue|default|do|double|else|extends|false|final|finally|float|for|function|goto|if|implements|import|in|instanceof|int|interface|long|native|new|null|package|private|protected|public|return|short|static|super|switch|synchronized|this|throw|throws|transient|true|try|var|void|while|with|\.appendChild|\.childNodes|\.createElement|\.getElementById|\.getElementsByName|\.getElementsByTagName|\.lastChild|\.nextSibling|\.parentNode|\.previousSibling|\.innerHTML|\.innerText|\.activeElement|\.alinkColor|\.all|\.bgColor|\.charset|\.clear|\.clear(selection)|\.close|\.cookie|\.createElement|\.createRange|\.createStyleSheet|\.defaultCharset|\.document|\.domain|\.elementFromPoint|\.embeds|\.empty|\.expando|\.fgColor|\.getSelection|\.handleEvent|\.lastModified|\.linkColor|\.open|\.parentWindow|\.readyState|\.referrer|\.releaseEvents|\.routeEvent|\.selection|\.showHelp|\.title|\.type|\.URL|\.vlinkColor|\.write|\.writeln|\.className|\.parentElement|\.tagName|\.style|\.action|\.blur|\.checked|\.click|\.defaultChecked|\.defaultSelected|\.defaultValue|\.elements|\.encoding|\.focus|\.form|\.forms|\.handleEvent|\.length|\.length|\.length|\.length|\.method|\.name|\.name|\.options|\.reset|\.select|\.selected|\.selectedIndex|\.submit|\.target|\.text|\.type|\.value|\.value|\.prototype)';
}
sub syntax
{
	my $class = shift;
	my $type = shift;
	return '>|js|\\nstring\\n||<' if ($type eq 'range');
	return;
}
# ======================================
package sbtext::Plugin::perl;
sub version
{
	return '0.02';
	# 0.02 [2008/12/04] ported for Serene Bach 3.*
	# 0.01 [2006/08/01] modified keyword and parse variables, quote, and comments
	# 0.00 [2006/07/19] generated
}
sub range
{
	my $class = shift;
	my $param = shift;
	my $obj = shift;
	my $keyword = $class->_keyword;
	$param =~ s/(\".*?\")/<span class="quote">$1<\/span>/g;
	$param =~ s/(\'.*?\')/<span class="quote">$1<\/span>/g;
	$param =~ s/$keyword/<em class="keyword">$1<\/em>/g;
	$param =~ s/(\$[A-Za-z_][A-Za-z0-9_>\-]*)/<em class="variable">$1<\/em>/g;
	$param =~ s/([\%\@][A-Za-z_][A-Za-z0-9_]*)/<em class="variable">$1<\/em>/g;
	$param =~ s/(#\s.+)\n/<span class="comment">$1<\/span>\n/g;
	$param =~ s/(#\s.+)$/<span class="comment">$1<\/span>/;
	return '<pre><code class="perl">' . $param . '</code></pre>';
}
sub _keyword
{
	return '(write|while|warn|wantarray|wait|vec|values|utime|use|untie|unshift|unpack|unlink|unless|undef|umask|ucfirst|uc|truncate|true|times|time|tied|tie|telldir|tell|syswrite|system|sysseek|sysread|sysopen|syscall|symlink|substr|sub|study|stat|srand|sqrt|sprintf|split|splice|sort|socketpair|socket|sleep|sin|shutdown|shmwrite|shmread|shmget|shmctl|shift|setsockopt|setservent|setpwent|setprotoent|setpriority|setpgrp|setnetent|sethostent|setgrent|send|semop|semget|semctl|select|seekdir|seek|scalar|rmdir|rindex|rewinddir|reverse|reverse|return|return|reset|require|rename|ref|redo|recv|readpipe|readlink|readline|readdir|read|rand|quotemeta|push|prototype|printf|print|pos|pop|pipe|package|pack|our|ord|opendir|open|oct|no|next|my|msgsnd|msgrcv|msgget|msgctl|mkdir|map|lstat|log|localtime|local|listen|link|length|lcfirst|lc|last|kill|keys|join|ioctl|int|index|import|if|hex|grep|goto|gmtime|glob|getsockopt|getsockname|getservent|getservbyport|getservbyname|getpwuid|getpwnam|getpwent|getprotoent|getprotobynumber|getprotobyname|getpriority|getppid|getpgrp|getpeername|getnetent|getnetbyname|getnetbyaddr|getlogin|gethostent|gethostbyname|gethostbyaddr|getgrnam|getgrgid|getgrent|getc|formline|format|fork|foreach|for|flock|fileno|fcntl|false|exp|exit|exists|exec|eval|eof|endservent|endpwent|endprotoent|endnetent|endhostent|endgrent|elsif|else|each|dump|do|die|delete|defined|dbmopen|dbmclose|crypt|cos|continue|connect|closedir|close|chroot|chown|chop|chomp|chmod|chdir|caller|bless|binmode|bind|atan2|alarm|accept|abs)';
}
sub syntax
{
	my $class = shift;
	my $type = shift;
	return '>|perl|\\nstring\\n||<' if ($type eq 'range');
	return;
}
# ======================================
package sbtext::Plugin::ruby;
sub version
{
	return '0.01';
	# 0.01 [2008/12/04] ported for Serene Bach 3.* / added range handler
	# 0.00 [2005/11/18] generated
}
sub range
{
	my $class = shift;
	my $param = shift;
	my $obj = shift;
	my $keyword = $class->_keyword;
	$param =~ s/(\".*?\")/<span class="quote">$1<\/span>/g;
	$param =~ s/(\'.*?\')/<span class="quote">$1<\/span>/g;
	$param =~ s/$keyword/<em class="keyword">$1<\/em>/g;
	$param =~ s/(#\s.+)\n/<span class="comment">$1<\/span>\n/g;
	return '<pre><code class="ruby">' . $param . '</code></pre>';
}
sub inline
{
	my $class = shift;
	my $param = shift;
	my $obj = shift;
	my ($string,$ruby) = split(',', $param);
	return qq(<ruby><rb>$string</rb><rp>(</rp><rt>$ruby</rt><rp>)</rp></ruby>);
}
sub _keyword
{
	return '(BEGIN|class|ensure|nil|self|when|END|def|false|not|super|while|alias|define|defined|for|or|then|yield|and|do|if|redo|true|begin|else|in|rescue|undef|break|elsif|module|retry|unless|case|end|next|return|until)';
}
sub syntax
{
	my $class = shift;
	my $type = shift;
	return '&ruby(string,ruby)' if ($type eq 'inline');
	return '>|ruby|\\nstring\\n||<' if ($type eq 'range');
	return;
}
# ======================================
package sbtext::Plugin::style;
sub version
{
	return '0.02';
	# 0.02 [2008/12/04] ported for Serene Bach 3.*
	# 0.01 [2006/03/17] changed bracket rule
	# 0.00 [2005/11/18] generated
}
sub inline
{ # e.g. 
	my $class = shift;
	my $param = shift;
	my $obj = shift;
	my ($color,$size,$string) = split(',', $param, 3);
	my $style = '';
	$style .= qq(color: $color;) if ($color =~ /^#[0-9a-f]{3,6}$/ or $color =~ /^\w+$/);
	if ($size =~ /^\d{1,3}(%|px)$/)
	{
		$style .= qq(font-size: $size;)
	}
	elsif ($size =~ /^(\+|-)([1-5])$/)
	{
		$size = ($1 eq '-') ? 100 - (10 * $2) : 100 + (10 * $2);
		$style .= qq(font-size: $size%;);
	}
	return ($style ne '') ? qq(<span style="$style">$string</span>) : qq(<span>$string</span>);
}
sub syntax
{
	my $class = shift;
	my $type = shift;
	return '&style(color,size,string)' if ($type eq 'inline');
	return;
}
1;
__END__
