# sb::Plugin::sbTextFormat
# == written by T.Otani <takuya.otani@gmail.com> ===
# == Copyright (C) 2008 SimpleBoxes/SerendipityNZ ==

# This plugin is distributed on the Perl Artistic License.
# ref: http://www.perl.com/pub/a/language/misc/Artistic.html

use strict;

if ( !defined $sb::VERSION
  or $sb::VERSION =~ /^2.(\d\d)[BRD]$/i )
{
	die("This plugin works under Serene Bach 3.* only.\n");
}

# ======================================
package sb::Plugin::sbTextFormat;
use sb::Plugin ();
# --------------------------------------
sb::Plugin->register_plugin(
	'lang' => {
		'ja' => 'utf8',
		'en' => 'ascii',
	},
	'text' => {
		'type'    => 'text format',
		'name'    => 'sbtext',
		'text'    => 'sbtext',
		'author'  => 'takkyun',
		'detail'  => 'http://serenebach.net/',
		'version' => version(),
	},
	'file' => undef,
	'data' => undef,
);
sb::Plugin->register_text_filter('sbtext');
sub version
{
	return '0.14';
	# 0.14 [2009/01/26] modified initialize and _desiding to handle heading properly
	# 0.13 [2008/12/06] fixed module name to work correctly
	# 0.12 [2008/11/24] refactored for Serene Bach 3.*
	# 0.11 [2007/04/12] changed get_footnote to fix a bug
	# 0.10 [2006/08/03] changed detail to point new site address
}
# ======================================
package sb::Text::sbtext;
use sb::Text::Base ();
use base qw( sb::Text::Base );
# --------------------------------------
# // public functions - class methdos
sub sbtext
{
	my $class  = shift;
	my $text   = shift;
	my $option = shift;
	my $parser = sb::Plugin::sbTextParser->get();
	$parser->prepare( $option );
	return $parser->parse( $text );
}
# ======================================
package sb::Plugin::sbTextParser;
use sb::Object ();
use sb::Data ();
use base qw( sb::RequestObject );
# --------------------------------------
# // defines / static variables
my $aRegLink = 's?https?:\/\/[-_.!~*\'()a-zA-Z0-9;\/?:\@&=+\$,%#]+';
my $aRegMail = 'mailto:[\w=+\$%*-]+\@[^\s()\[\]{}!\"\'<>:,\x7f-\xff]+\.\w+';
# --------------------------------------
# // override base class
sub initialize
{
	my $self = shift;
	$self->SUPER::initialize(
		'link_target'     => ' target="_blank"',                               # attribute for external link
		'heading_mark'    => '&#9632;',                                        # mark string of heading
		'heading_prefix'  => 'eid',                                            # id prefix of heading
		'lowest_heading'  => 2,                                                # lowest(biggest) level of heading
		'max_block_level' => 3,                                                # maximum quote level for Hatena block
		'footnote_prefix' => 'note',                                           # id prefix of footnote
		'note_begin_tag'  => '<sup>',                                          # beginning tag of note mark
		'note_end_tag'    => '</sup>',                                         # end tag of note mark
		'link_hatena'     => 'http://d.hatena.ne.jp/keyword/',                 # hatena link
		'link_google'     => 'http://www.google.com/search?lr=lang_ja&amp;q=', # search link
		'link_amazon'     => 'http://www.amazon.co.jp/exec/obidos/ASIN/',      # amazon link
		'link_wikipedia'  => 'http://ja.wikipedia.org/wiki/',                  # wikipedia link
		'amazon_id'       => 'simpleboxes-22',                                 # amazon associate id
		'baseurl'         => '',                                               # baseurl for inner link
		'plugin'          => sbtext::PluginManager->new(),                     # sbtext plugin manager
		'keyword'         => [],                                               # keywords to autolink
		'empty_tag'       => [],                                               # allowed empty elements
		'inline_tags'     => [],                                               # allowed inline elements
		'cat'             => undef,                                            # article categories
		'cache'           => {},                                               # [variable] result cache
		'footnotes'       => {},                                               # [variable] buffer of footnotes
		'heading'         => {},                                               # [variable] buffer of headings
		'inner_keywords'  => [],                                               # [variable] buffer of included keywords
		'userid'          => undef,                                            # [variable] id of author who edits the target entry
		'mode'            => '',                                               # [argument] current mode (body / more / body_only)
		'obj'             => undef,                                            # [argument] target entry object
	);
	$self->{'empty_tag'} = ['hr','br','img'];
	$self->{'inline_tags'} = ['a','q','strong','em','abbr','code','del','ins','p','div'];
	return $self;
}
# --------------------------------------
# // public functions
sub userid
{
	my $self = shift;
	if ( @_ )
	{
		my $user = shift;
		if ( $user ne '' and $user ne $self->{'userid'} )
		{
			$self->{'userid'} = $user;
			$user = sb::Data->load('User','id'=>$self->{'userid'});
			$self->{'amazon_id'} = $user->aws if ( $user and $user->aws ne '' );
		}
		else
		{
			$self->{'userid'} = '';
			$self->{'amazon_id'} = '';
		}
		$self->{'amazon_id'} = 'simpleboxes-22' if ( $self->{'amazon_id'} eq '' );
	}
	return $self->{'userid'};
}
sub obj
{
	my $self = shift;
	if ( @_ )
	{
		my $obj = shift;
		$obj = undef if ( ref($obj) ne 'sb::Data::Entry' );
		if ( !$obj
		  or !$self->{'obj'}
		  or $obj->id ne $self->{'obj'}->id )
		{
			$self->userid( $obj ? $obj->auth : '' );
			$self->{'obj'} = $obj;
			$self->{'footnotes'} = {'body'=>[],'more'=>[]};
			$self->{'heading'} = {'body'=>[],'more'=>[]};
			$self->{'cache'} = {};
			if ( $self->mode eq 'body' and $self->{'obj'}->more ne '' )
			{
				$self->{'mode'} = 'more';
				$self->parse( $self->{'obj'}->more );
				$self->{'mode'} = 'body';
			}
		}
	}
	return $self->{'obj'};
}
sub mode
{
	my $self = shift;
	if ( @_ )
	{
		my $mode = shift;
		$mode = 'body' if ( $mode ne 'more' and $mode ne 'body_only' );
		$self->{'mode'} = $mode;
	}
	return $self->{'mode'} 
}
sub cat
{
	my $self = shift;
	$self->{'cat'} = {sb::Data->load_as_hash('Category')} if ( !$self->{'cat'} );
	return $self->{'cat'};
}
sub prepare
{
	my $self  = shift;
	my $param = shift;
	if ( ref($param) eq 'HASH' )
	{
		$self->mode($param->{'mode'}); # body or more or body_only
		$self->obj($param->{'obj'});   # an instance of entry
	}
	else
	{
		$self->mode('body');
		$self->obj(undef);
	}
}
sub parse
{
	my $self = shift;
	my $text = shift;
	return $text if ( !$text );
	return $self->{'cache'}->{$self->mode} if ( $self->{'cache'}->{$self->mode} );
	$text = $self->_starting($text);  # initialize
	$text = $self->_siding($text);    # side special letters
	$text = $self->_makenote($text);  # make footnotes (buffering)
	$text = $self->_range($text);     # handle range markers
	$text = $self->_blocks($text);    # parse blocks
	$text = $self->_desiding($text);  # deside special letters
	$text = $self->_finishing($text); # tidy up
	$text = $self->_add_note($text);  # add footnotes (debuffering)
	$self->{'cache'}->{$self->mode} = $text;
	return $text;
}
# ------------------------------------------------
# // private functions - parser
sub _starting
{
	my $self = shift;
	my $text = &_entitize(shift);
	my @inner = ();
	foreach my $key ( @{$self->{'keyword'}} )
	{
		push(@inner,$key) if ( index($text,$key) > -1 );
	}
	$self->{'inner_keywords'} = \@inner;
	return &_linefeed($text);
}
sub _finishing
{
	my $self = shift;
	my $text = shift;
	$text =~ s!<p>\n!<p>!g;
	$text =~ s!<p (.*?)>\n!<p $1>!g;
	$text =~ s!<pre>\n!<pre>!g;
	$text =~ s!\n</pre>!</pre>!g;
	$text =~ s!<p></p>\n!!g;
	$text =~ s!<br />\n</p>!</p>!g;
	return $text;
}
sub _siding
{
	my $self = shift;
	my $text = shift;
	$text =~ s/\\\\/&sb_;/g;
	$text =~ s/\\\^/&sba;/g;
	$text =~ s/\\\*/&sbb;/g;
	$text =~ s/\\\&lt;/&sbc;/g;
	$text =~ s/\\\'/&sbd;/g; # escape '
	$text =~ s/\\\(/&sbe;/g;
	$text =~ s/\\\[/&sbf;/g;
	$text =~ s/\\\|/&sbg;/g;
	$text =~ s/\\\-/&sbh;/g;
	$text =~ s/\\\+/&sbi;/g;
	$text =~ s/\\\:/&sbj;/g;
	$text =~ s/\\\#/&sbk;/g;
	$text =~ s/\\\)/&sbl;/g;
	$text =~ s/\\\]/&sbm;/g;
	$text =~ s/\\\&amp;/&sbn;/g;
	$text =~ s/(^|\n)#contents.*?(\n|$)/$1<sbcontents \/>\n/g;
	$text =~ s/(^|\n)##/$1&sbk;#/g;
	return $text;
}
sub _desiding
{
	my $self = shift;
	my $text = shift;
	my @buffer = ();
	my $permalink = '';
	if ( $self->obj )
	{
		$permalink = $self->obj->permalink('cat'=>$self->cat);
	}
	foreach my $mode ( 'body','more' )
	{
		next if ( !@{$self->{'heading'}->{$mode}} );
		foreach ( @{$self->{'heading'}->{$mode}} )
		{
			s/\[(.*?)\:#(.*?)\]/[$1\:$permalink\#$2]/;
			push( @buffer, $_ );
		}
	}
	my $heading = $self->_blocks(join("\n",@buffer));
	$text =~ s/<sbcontents \/>/$heading/g;
	$text =~ s/\&sbn;/\&amp;/g;
	$text =~ s/\&sbm;/\]/g;
	$text =~ s/\&sbl;/\)/g;
	$text =~ s/\&sbk;/\#/g;
	$text =~ s/\&sbj;/\:/g;
	$text =~ s/\&sbi;/\+/g;
	$text =~ s/\&sbh;/\-/g;
	$text =~ s/\&sbg;/\|/g;
	$text =~ s/\&sbf;/\[/g;
	$text =~ s/\&sbe;/\(/g;
	$text =~ s/\&sbd;/\'/g; # unescape '
	$text =~ s/\&sbc;/\&lt;/g;
	$text =~ s/\&sbb;/\*/g;
	$text =~ s/\&sba;/\^/g;
	$text =~ s/\&sb_;/\\/g;
	return $text;
}
sub _makenote
{
	my $self = shift;
	my $text = shift;
	my $mode = $self->mode;
	$mode eq 'body' if ( $mode eq 'body_only' );
	if ( $self->{'footnotes'}->{$mode}
	 and @{$self->{'footnotes'}->{$mode}} )
	{
		$text =~ s/\(\((.*?)\)\)/<sbnote \/>/g;
	}
	else
	{
		$self->{'footnotes'}->{$mode} = [];
		push(@{$self->{'footnotes'}->{$mode}},$1) while ( $text =~ s/\(\((.*?)\)\)/<sbnote \/>/ );
	}
	return $text;
}
sub _add_note
{
	my $self = shift;
	my $text = shift;
	my $mode = $self->mode;
	$mode = 'body' if ( $mode eq 'body_only' );
	my $id = ( $self->obj ) ? $self->obj->id : '';
	my @notes = @{$self->{'footnotes'}->{'body'}};
	my $notes_body = @notes;
	push(@notes,@{$self->{'footnotes'}->{'more'}}) if ( $mode eq 'more' );
	if (@notes)
	{
		my $footnote = '<ul class="footnote">';
		my $bgn = $self->{'note_begin_tag'};
		my $end = $self->{'note_end_tag'};
		my $symbol = $self->{'footnote_prefix'} . $id . '-';
		for (my $i = 0; $i < @notes; $i++)
		{
			my $note = sb::Text->new($self->_inline($notes[$i]));
			my $note_foot = $note->data;
			my $note_body = $note->tag_removed;
			my $footnum = $i + 1;
			my $bodynum = $footnum;
			my $foot = $symbol . $i . '-body';
			my $body = $symbol . $i;
			if ( $mode eq 'more' )
			{ # we need to set offset for notes in body
				$bodynum += $notes_body;
				$note_body = sb::Text->new($self->_inline($notes[$i + $notes_body]))->tag_removed;
			}
			$text =~ s!<sbnote />!$bgn<a href="#$foot" id="$body" title="$note_body">*$bodynum </a>$end!;
			$footnote .= qq(\n<li id="$foot"><a href="#$body">*$footnum</a> : $note_foot</li>);
		}
		$footnote .= "\n</ul>\n";
		$text .= $footnote if ( !$self->obj or $self->mode ne 'body' or $self->obj->more eq '');
	}
	return $text;
}
sub _range
{
	my $self = shift;
	my $text = shift;
	my @result = ();
	my $quote = -1;
	my $pre   = -1;
	$quote = 0 if ($text =~ /&gt;&gt;(.*)\n&lt;&lt;/s);
	$pre   = 0 if ($text =~ /&gt;\|\|\n(.*)\n\|\|&lt;/s);
	while ($text =~ /&gt;\|(\w+)\|\n(.*?)\n\|\|&lt;/s)
	{
		my $name  = $1;
		my $value = $2;
		$value =~ s/\\/&sb_;/g;
		$value =~ s/\n/\\n/g;
		$text =~ s/&gt;\|(\w+)\|\n(.*?)\n\|\|&lt;/##$name\($value\)\n/s;
	}
	return $text if ($quote == -1 and $pre == -1);
	my @buf = split("\n",$text);
	foreach my $line (@buf)
	{
		if ($line =~ /^&lt;&lt;$/)
		{
			$quote--;
			$quote = 0 if ($quote < 0);
			next;
		}
		elsif ($line =~ /^\|\|&lt;$/)
		{
			$pre = 0;
			next;
		}
		if ($quote >= 0 and $line =~ /^&gt;&gt;(.*)$/)
		{
			$quote++;
			$quote = $self->{'max_block_level'} if ($quote > $self->{'max_block_level'});
			$line = $1;
			next if ($line eq '');
		}
		elsif ($pre >= 0 and $line =~ /^&gt;\|\|/)
		{
			$pre = 1;
			next;
		}
		my $mark = ($quote > 0) ? '&gt;' x $quote
		         : ($pre > 0) ? ' ' 
		         : '';
		push(@result,$mark . $line);
	}
	return join("\n",@result);
}
sub _blocks
{ # from YukiWiki / Copyright (C) 2000-2004 Hiroshi Yuki <hyuki@hyuki.com>
	my $self = shift;
	my $text = shift;
	my (@result,@saved);
	my $heading = 0;
	my $quote_flag = 0;
	my @buf = split("\n",$text);
	unshift(@saved, '</p>');
	push(@result, '<p>');
	foreach (@buf)
	{
		my $max_level = $self->{'max_block_level'};
		if (/^(\*{1,$max_level})(.+)/)
		{ # heading
			my ($hn,$id,$mark,$list);
			my $number = (length($1) + $self->{'lowest_heading'} - 1);
			my $label = sb::Text->new($self->_inline($2));
			$number = 6 if ($number > 6);
			$hn = 'h' . $number;
			$id = $self->{'heading_prefix'};
			$id .= $self->obj->id if ( $self->obj );
			$id .= '-' . $self->mode . $heading;
			$mark = $self->{'heading_mark'};
			$mark = '<a href="#' . $id . '">' . $mark . '</a>' if ($mark ne '');
			$list = $label->tag_removed;
			$list =~ s/\:/&sbj;/g;
			$list =~ s/\#/&sbk;/g;
			push(@result, splice(@saved));
			push(@result, qq(<$hn id="$id">$mark) . $$label  . qq(</$hn>));
			unshift(@saved, '</p>');
			push(@result, '<p>');
			if ( $self->mode ne 'body_only' )
			{
				push(@{$self->{'heading'}->{$self->mode}},('-' x $number) . '[' . $list  . ':#' .  $id . ']');
			}
			$heading++;
		}
		elsif (/^====/)
		{ # horizontal rule
			push(@result, splice(@saved), '<hr />');
		}
		elsif (/^----/ and $max_level < 4)
		{ # horizontal rule (backword-compatibility)
			push(@result, splice(@saved), '<hr />');
		}
		elsif (/^\/\/(.+)/)
		{ # note
			&_back_push(
				'tag'       => 'p',
				'level'     => 1,
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => ' class="note"',
			);
			push(@result, $self->_inline($_) . '<br />');
		}
		elsif (/^(-{1,$max_level})(.+)/)
		{ # unordered list
			&_back_push(
				'tag'       => 'ul',
				'level'     => length($1),
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => '',
			);
			push(@result, '<li>' . $self->_inline($2) . '</li>');
		}
		elsif (/^(\+{1,$max_level})(.+)/)
		{ # ordered list
			&_back_push(
				'tag'       => 'ol',
				'level'     => length($1),
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => '',
			);
			push(@result, '<li>' . $self->_inline($2) . '</li>');
		}
		elsif (/^:([^:]+):(.+)/)
		{ # definition list
			&_back_push(
				'tag'       => 'dl',
				'level'     => 1,
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => '',
			);
			push(@result, '<dt>' . $self->_inline($1) . '</dt>', '<dd>' . $self->_inline($2) . '</dd>');
		}
		elsif (/^((&gt;){1,$max_level})(.*)/)
		{ # quote block
			my $attribute = '';
			my $level = length($1) / 4;
			my $quote = $3;
			if ($quote =~ /^=(.*?):($aRegLink)?/)
			{
				$attribute .= ' title="' . $1 . '"' if ($1 ne '');
				$attribute .= ' cite="' . $2 . '"' if ($2 ne '');
			}
			&_back_push(
				'tag'       => 'blockquote',
				'level'     => $level,
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => $attribute,
			);
			next if ($attribute ne '');
			push(@result, '<p>' . $self->_inline($quote) . '</p>' . "\n");
		}
		elsif (/^$/)
		{ # empty line (end of paragraph)
			push(@result, splice(@saved));
			unshift(@saved, '</p>');
			push(@result, '<p>');
		}
		elsif (/^(\s+?)(.*)$/)
		{ # preformatted text
			&_back_push(
				'tag'       => 'pre',
				'level'     => 1,
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => '',
			);
			push(@result, $2);
		}
		elsif (/^\|(.*?)$/)
		{ # table
			my $tmp = $1;
			&_back_push(
				'tag'       => 'table',
				'level'     => 1,
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => '',
			);
			my @elems = ($tmp =~ /(.*?)\|/g);
			my $value = '';
			foreach my $elem (@elems)
			{
				if ($elem =~ /^\*(.*)/)
				{
					$value .= '<th>' . $1 . '</th>'
				}
				else
				{
					$value .= '<td>' . $elem . '</td>';
				}
			}
			push(@result, '<tr>' . $value . '</tr>');
		}
		elsif (/^\&lt;(p|div)(.*?)\&gt;(.+?)\&lt;\/(p|div)\&gt;$/)
		{ # allowed block elements
			&_back_push(
				'tag'       => $1,
				'level'     => 1,
				'saved'     => \@saved,
				'result'    => \@result,
				'attribute' => &_sanitize_attribute($2),
			);
			push(@result, $self->_inline($3));
		}
		elsif (/^#(\w+)(\((.*)\))?/)
		{ # block plugin
			my $original = $_;
			my $result = ($self->{'plugin'}) ? $self->{'plugin'}->block($self,$1,$3) : undef;
			if (defined($result))
			{
				push(@result, splice(@saved));
			}
			else
			{
				$result = $original;
			}
			push(@result, $result);
		}
		elsif (/^##(\w+)(\((.*)\))?/)
		{ # range plugin
			my $tag = $1;
			my $value = $3;
			my $result = undef;
			$value =~ s/\\n/\n/g;
			$result = $self->{'plugin'}->range($self,$tag,$value) if ($self->{'plugin'});
			if (defined($result))
			{
				push(@result, splice(@saved));
				push(@result, $result);
			}
			else
			{
				push(@result,"<p>&gt;&sbg;$tag&sbg;<br />");
				foreach my $line ( split("\n",$value) )
				{
					push(@result,$self->_inline($line) . '<br />');
				}
				push(@result,"&sbg;&sbg;&lt;</p>");
			}
		}
		else
		{ # paragraph
			push(@result, $self->_inline($_) . '<br />');
		}
	}
	push(@result, splice(@saved));
	return join("\n",@result) . "\n";
}
sub _inline
{
	my $self = shift;
	my $text = shift;
	$text =~ s!'''(.*?)'''!<em>$1</em>!g;           # emphasis (wiki syntax)
	$text =~ s!''(.*?)''!<strong>$1</strong>!g;     # strong emphasis (wiki syntax)
	$text =~ s!\*\*(.*?)\*\*!<strong>$1</strong>!g; # strong emphasis
	$text =~ s!\*(.*?)\*!<em>$1</em>!g;             # emphasis
	foreach my $key ( @{$self->{'inner_keywords'}} )
	{ # preparation for keywords
		$text =~ s/(\Q$key\E)/[[$1]]/g;
	}
	# $text =~ s!\[\[(.*?)\]\]!<a href="$self->{'baseurl'}/$1" class="keyword">$1</a>!g; # keywords
	while ($text =~ /\^(.*?)\((.*?)\)/)
	{ # ruby and link
		my $check = $2;
		if ($check =~ /($aRegLink)/)
		{
			my $target = (index($check,$self->{'baseurl'}) > -1) ? '' : $self->{'link_target'};
			$text =~ s!\^(.*?)\((.*?)\)!<a href="$2"$target>$1</a>!;
		}
		elsif ($check =~ /($aRegMail)/)
		{
			$text =~ s!\^(.*?)\((.*?)\)!<a href="$2">$1</a>!;
		}
		else
		{
			$text =~ s!\^(.*?)\((.*?)\)!<ruby><rb>$1</rb><rp>(</rp><rt>$2</rt><rp>)</rp></ruby>!;
		}
	}
	while ($text =~ /\[(.*?):(.*?)\]/)
	{ # hatena, google, link
		my ($name,$check) = ($1,$2);
		if ($name ne '' and ($check =~ /($aRegLink)/ or $check =~ /^#/))
		{
			my $target = ($check =~ /^#/ and index($check,$self->{'baseurl'}) > -1) ? '' : $self->{'link_target'};
			$text =~ s!\[(.*?):(.*?)\]!<a href="$2"$target>$name</a>!;
		}
		else
		{
			my $target = $self->{'link_target'};
			my $amazon = $self->{'link_amazon'};
			my $hatena = $self->{'link_hatena'};
			my $google = $self->{'link_google'};
			my $wpedia = $self->{'link_wikipedia'};
			if ($name =~ /ASIN/i and $check =~ /^\w+$/)
			{ # Amazon
				$check = $amazon . $check . '/' . $self->{'amazon_id'};
				$text =~ s/\[(ASIN):(.*?)\]/<a href="$check"$target>$1:$2<\/a>/i;
			}
			elsif ($name =~ /hatena/i)
			{ # keyword for Hatena
				$check = sb::Language->get->convert($check,'euc');
				$check =~ s/(\W)/'%' . unpack('H2', $1)/eg;
				$text =~ s/\[(.*?):(.*?)\]/<a href=\"$hatena$check\"$target>$2<\/a>/;
			}
			elsif ($name =~ /google/i)
			{ # word for Google
				$check = sb::Language->get->convert($check,'utf8');
				$check =~ s/(\W)/'%' . unpack('H2', $1)/eg;
				$text =~ s/\[(.*?):(.*?)\]/<a href=\"$google$check\"$target>$2<\/a>/;
			}
			elsif ($name =~ /wikipedia/i)
			{ # word for wikipedia
				$check = sb::Language->get->convert($check,'utf8');
				$check =~ s/(\W)/'%' . unpack('H2', $1)/eg;
				$text =~ s/\[(.*?):(.*?)\]/<a href=\"$wpedia$check\"$target>$2<\/a>/;
			}
			else
			{ # escape blackets
				$text =~ s/\[(.*?):(.*?)\]/&sbf;$1\:$2\]/;
			}
		}
	}
	foreach my $tag ( @{$self->{'empty_tag'}} )
	{ # allowed empty elements
		while ($text =~ /\&lt;$tag\s(.*?)\/\&gt;/i)
		{
			my $attr = &_sanitize_attribute($1);
			$text =~ s!\&lt;$tag\s(.*?)/\&gt;!<$tag $attr/>!i;
		}
	}
	foreach my $tag ( @{$self->{'inline_tags'}} )
	{ # allowed inline elements
		while ($text =~ /\&lt;$tag\s(.*?)\&gt;(.+?)\&lt;\/$tag\&gt;/i)
		{
			my $attr = &_sanitize_attribute($1);
			my $elem = $2;
			$text =~ s/\&lt;$tag\s(.*?)\&gt;(.+?)\&lt;\/$tag\&gt;/<$tag $attr>$elem<\/$tag>/i;
		}
	}
	if ($self->{'plugin'} and $text =~ /\&amp;(\w+)\((([^()]*(\([^()]*\))?)*)\)/)
	{ # inline plugin
		$text =~ s/\&amp;(\w+)\((([^()]*(\([^()]*\))?)*)\)/&_inline_plugin($self,$1,$2)/eg;
	}
	if ($text =~ /($aRegLink)/ and $text !~ /"$aRegLink"/)
	{ # inline uri => http
		my $target = (index($1,$self->{'baseurl'}) > -1) ? '' : $self->{'link_target'};
		$text =~ s/($aRegLink)/<a href=\"$1\"$target>$1<\/a>/g;
	}
	if ($text =~ /($aRegMail)/ and $text !~ /"$aRegMail"/)
	{ # inline uri => mailto
		$text =~ s/($aRegMail)/<a href="$1">$1<\/a>/g;
	}
	return $text;
}
sub _inline_plugin
{
	my $self = shift;
	my $name = shift;
	my $param = shift;
	my $result = $self->{'plugin'}->inline($self,$name,$param);
	return (defined($result)) ? $result : '&amp;' . $name . '(' . $param . ')';
}
# ----------------------------------------------------
# // private functions - utilities
sub _back_push
{ # from YukiWiki / Copyright (C) 2000-2004 Hiroshi Yuki <hyuki@hyuki.com>
	my %param = (
		'tag'       => undef,
		'level'     => undef,
		'saved'     => undef,
		'result'    => undef,
		'attribute' => undef,
		@_
	);
	my $bgn_tag = '<' . $param{'tag'} . $param{'attribute'} . '>';
	my $end_tag = '</' . $param{'tag'} . '>';
	if ($param{'tag'} ne 'blockquote' and $param{'attribute'} ne '')
	{
		$param{'attribute'} =~ s/\-/_/g;
		$end_tag .= '<!--' . $param{'attribute'} . '-->';
	}
	while (@{$param{'saved'}} > $param{'level'})
	{
		push(@{$param{'result'}}, shift(@{$param{'saved'}}));
	}
	if ($param{'saved'}->[0] ne $end_tag)
	{
		push(@{$param{'result'}}, splice(@{$param{'saved'}}));
	}
	while (@{$param{'saved'}} < $param{'level'})
	{
		unshift(@{$param{'saved'}}, $end_tag);
		push(@{$param{'result'}}, $bgn_tag);
	}
}
sub _entitize
{
	my $text = shift;
	$text =~ s/&/&amp;/g;
	$text =~ s/\"/&quot;/g;
	$text =~ s/</&lt;/g;
	$text =~ s/>/&gt;/g;
	$text =~ s/\{/&#123;/g;
	$text =~ s/\}/&#125;/g;
	return $text;
}
sub _detitize
{
	my $text = shift;
	$text =~ s/&#123;/\{/g;
	$text =~ s/&#125;/\}/g;
	$text =~ s/&gt;/>/g;
	$text =~ s/&lt;/</g;
	$text =~ s/&quot;/\"/g;
	$text =~ s/&amp;/&/g;
	return $text;
}
sub _linefeed
{
	my $text = shift;
	$text =~ s/\x0D\x0A/\n/g;
	$text =~ tr/\x0D\x0A/\n\n/;
	return $text;
}
sub _sanitize_attribute
{
	my $text = &_detitize(shift);
	my @orig = ($text =~ /(\w+?=\".*?\")/g);
	my @attr = ();
	foreach my $check (@orig)
	{
		my ($name,$val) = split('=',$check,2);
		$name =~ s/\s//g;
		push(@attr,$name . '=' . &_remove_space($val)) if ($name =~ /id|class|lang|dir|title|href|target|cite/i);
	}
	return( ' ' . join(' ',@attr) );
}
sub _remove_space
{
	my $text = shift;
	$text =~ s/^\s*//g; # removed most left spaces
	$text =~ s/\s*$//g; # removed most right spaces
	return $text;
}
# ======================================
package sbtext::PluginManager;
use sb::Object ();
use base qw( sb::RequestObject );
# --------------------------------------
# // override base class
sub initialize
{
	my $self = shift;
	return $self;
}
# --------------------------------------
# // public functions - handlers
sub block
{
	my $self = shift;
	return &_handle('block',@_);
}
sub inline
{
	my $self = shift;
	return &_handle('inline',@_);
}
sub range
{
	my $self = shift;
	return &_handle('range',@_);
}
sub syntax
{
	my $self = shift;
	my $name = shift;
	return &_handle('syntax',undef,$name,@_);
}
# --------------------------------------
# // private functions
sub _handle
{
	my $type   = shift;
	my $parser = shift;
	my $name   = shift;
	my $param  = shift;
	my $obj = ($parser) ? $parser->obj : undef;
	my $result = undef;
	eval {
		$result = &_name($name)->$type($param,$obj);
	};
	return ($@) ? undef : $result;
}
sub _name
{
	return 'sbtext::Plugin::' . $_[0];
}
1;
__END__
